#!/usr/bin/python
# -*- coding: utf-8 -*-

#***************************** License *************************************
#
#    Copyright (C) 2014  Royer Amélie
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#****************************************************************************

""" **generate_task_sequence_KS.py**

Generates a labels sequence from a Kernelized sorting grid.
"""

__author__= "Amélie Royer"
__date__ = "2014"
__email__= "amelie.royer@ist.ac.at"


import os
import sys
import numpy as np
from random import randrange, choice, random
from data_representation import kernelized_sorting_map




def get_neighbour(x, y, pos, xlim, ylim):
    """
    Returns the neighbour at position "pos" of the cell (x,y) in the grid.

    Args:
     * ``x`` (*int*): absciss.
     * ``y`` (*int*): ordinate..
     * ``pos`` (*int*): determines which neighbour to return (0 for bottom, 1 for right, 2 for top, 3 for left, 4 for center).
     * ``xlim`` (*int*): length of the grid on x-axis.
     * ``ylim`` (*int*): length of the grid on y-axis.

    Returns:
     * ``n`` (*tuple*): the chosen neighbout of the cell (x,y).
    """

    if pos == 0:
        return (x-1) % xlim, y
    elif pos == 1:
        return x, (y+1)%ylim
    elif pos == 2:
        return (x+1)%xlim, y
    elif pos == 3:
        return x, (y-1)%ylim
    elif pos == 4:
        return x, y
    else:
        return None



def random_walk_grid(n, grid, tree_struct, height=0, filename=None, hypernims=None, lambd=0):
    """
    This function returns a sequence of labels computed with a random walk on a KS 2D grid-representation of the data.

    Args:
     * ``n`` (*int*): sequence's length.
     * ``grid`` (*ndarray*): KS 2D projection of the labels.
     * ``tree_struct`` (*dict*): contains the nodes of the hierarchy tree (see ``extract_data.Node``).
     * ``height`` (*int, optional*): If given, and filename and hypernims are not None, then the name and ID of the ancester of this height for each label will be output with the sequence.
     * ``filename`` (*str, optional*): path to the file the sequence will be written in (Imagenet ID). Defaults to None.
     * ``hypernims`` (*list, optional*): list mapping a leaf-label to one of its ancestors for a given height. These relations will be used to display the ancestors classes. Defaults to None.
     * ``lambd`` (*float*): probability of making a random jump.


    Returns:
     * ``label_list`` (*list*): a sequence of ``n`` labels. (Starting index 0).
    """

    xlim, ylim = grid.shape
    label_list = [0] * n
    if filename is not None:
        f = open(filename, 'w')

    # Random Walk
    x, y = randrange(xlim), randrange(ylim)
    for step in xrange(n):
        classID = grid[x,y] + 1 # Imagenet class
        class_name = tree_struct[classID].synset[0]

        if height != 0 and hypernims is not None:
            ancestorID = hypernims[classID][0]
            ancestor_name = tree_struct[ancestorID].synset[0]
            s = "%d | %s | %d | %d | %s | %d \n" %(classID, class_name, 0, ancestorID, ancestor_name, height)
        else:
            s = "%d | %s \n" %(classID, class_name)
        if filename is not None:
            f.write(s)

        label_list[step] = classID - 1

        #Get next
        jump = random()
        if jump < lambd:
            x, y = randrange(xlim), randrange(ylim)
        else:
            x, y = get_neighbour(x, y, randrange(5), xlim, ylim)



    if filename is not None:
        print 'Output saved in', filename
        f.close()

    return label_list



def get_sequence_KS(n, dists, tree_struct, imported_coordinates=None, height=None, hypernims=None, lambd=0, output_walk=None, output_grid_file=None):
    """
    This function builds a KS projection of the labels space from the given distance matrix, and then returns a sequence of labels obtained through a random walk on this space.

    Args:
     * ``n`` (*int*): sequence's length.
     * ``dists`` (*ndarray*): distance matrix for the projection.
     * ``tree_struct`` (*dict*): contains the nodes of the hierarchy tree (see ``extract_data.Node``).
     * ``imported_coordinates`` (*str, optional*): path to the file containing the precomputed projected datapoint coordinates. Defaults to None.
     * ``height`` (*int, optional*): See random_walk_grid.
     * ``hypernims`` (*list, optional*) : See random_walk_grid.
     * ``lambd`` (*float*): probability of making a random jump.
     * ``output_walk`` (*str, optional*): path to the file the sequence will be written in. Defaults to None.
     * ``output_grid_file`` (*str, optional*): path to the file the coordinates will be written in. Defaults to None.

    Returns:
     * ``label_list`` (*list*): a sequence of n labels.
    """

    if imported_coordinates is None or not os.path.isfile(imported_coordinates):
        patching = kernalized_sorting_map(dists, outputfile=output_grid_file)
    else :
        print "Import coordinates from ", imported_coordinates
        patching = np.loadtxt(imported_coordinates, dtype=np.int32)

    return random_walk_grid(n, patching, tree_struct, height=height, filename=output_walk, hypernims=hypernims, lambd=lambd)
